import logging
from abc import ABC
from base64 import b64decode
from collections import deque, defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import (Set, FrozenSet, Optional, Union, List,
                    DefaultDict, Annotated, Literal)
from uuid import UUID

import pytest

from dataclass_wizard.v0 import *
from dataclass_wizard.v0.class_helper import get_meta
from dataclass_wizard.v0.constants import TAG
from dataclass_wizard.v0.errors import ParseError
from ..conftest import *
from ..._typing import *


log = logging.getLogger(__name__)


def test_asdict_and_fromdict():
    """
    Confirm that Meta settings for both `fromdict` and `asdict` are merged
    as expected.
    """

    @dataclass
    class MyClass:
        my_bool: Optional[bool]
        myStrOrInt: Union[str, int]

    d = {'myBoolean': 'tRuE', 'my_str_or_int': 123}

    LoadMeta(
        key_transform='CAMEL',
        raise_on_unknown_json_key=True,
        json_key_to_field={'myBoolean': 'my_bool', '__all__': True}
    ).bind_to(MyClass)

    DumpMeta(key_transform='SNAKE').bind_to(MyClass)

    # Assert that meta is properly merged as expected
    meta = get_meta(MyClass)
    assert 'CAMEL' == meta.key_transform_with_load
    assert 'SNAKE' == meta.key_transform_with_dump
    assert True is meta.raise_on_unknown_json_key
    assert {'myBoolean': 'my_bool'} == meta.json_key_to_field

    c = fromdict(MyClass, d)

    assert c.my_bool is True
    assert isinstance(c.myStrOrInt, int)
    assert c.myStrOrInt == 123

    new_dict = asdict(c)

    assert new_dict == {'myBoolean': True, 'my_str_or_int': 123}


def test_asdict_with_nested_dataclass():
    """Confirm that `asdict` works for nested dataclasses as well."""

    @dataclass
    class Container:
        id: int
        submittedDt: datetime
        myElements: List['MyElement']

    @dataclass
    class MyElement:
        order_index: Optional[int]
        status_code: Union[int, str]

    submitted_dt = datetime(2021, 1, 1, 5)
    elements = [MyElement(111, '200'), MyElement(222, 404)]

    c = Container(123, submitted_dt, myElements=elements)

    DumpMeta(key_transform='SNAKE',
             marshal_date_time_as='TIMESTAMP').bind_to(Container)

    d = asdict(c)

    expected = {
        'id': 123,
        'submitted_dt': round(submitted_dt.timestamp()),
        'my_elements': [
            # Key transform now applies recursively to all nested dataclasses
            # by default! :-)
            {'order_index': 111, 'status_code': '200'},
            {'order_index': 222, 'status_code': 404}
        ]
    }

    assert d == expected


def test_tag_field_is_used_in_dump_process():
    """
    Confirm that the `_TAG` field appears in the serialized JSON or dict
    object (even for nested dataclasses) when a value is set in the
    `Meta` config for a JSONWizard sub-class.
    """

    @dataclass
    class Data(ABC):
        """ base class for a Member """
        number: float

    class DataA(Data):
        """ A type of Data"""
        pass

    class DataB(Data, JSONWizard):
        """ Another type of Data """
        class _(JSONWizard.Meta):
            """
            This defines a custom tag that shows up in de-serialized
            dictionary object.
            """
            tag = 'B'

    @dataclass
    class Container(JSONWizard):
        """ container holds a subclass of Data """
        class _(JSONWizard.Meta):
            tag = 'CONTAINER'

        data: Union[DataA, DataB]

    data_a = DataA(number=1.0)
    data_b = DataB(number=1.0)

    # initialize container with DataA
    container = Container(data=data_a)

    # export container to string and load new container from string
    d1 = container.to_dict()

    expected = {
        TAG: 'CONTAINER',
        'data': {'number': 1.0}
    }

    assert d1 == expected

    # initialize container with DataB
    container = Container(data=data_b)

    # export container to string and load new container from string
    d2 = container.to_dict()

    expected = {
        TAG: 'CONTAINER',
        'data': {
            TAG: 'B',
            'number': 1.0
        }
    }

    assert d2 == expected


def test_to_dict_key_transform_with_json_field():
    """
    Specifying a custom mapping of JSON key to dataclass field, via the
    `json_field` helper function.
    """

    @dataclass
    class MyClass(JSONSerializable):
        my_str: str = json_field('myCustomStr', all=True)
        my_bool: bool = json_field(('my_json_bool', 'myTestBool'), all=True)

    value = 'Testing'
    expected = {'myCustomStr': value, 'my_json_bool': True}

    c = MyClass(my_str=value, my_bool=True)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    assert result == expected


def test_to_dict_key_transform_with_json_key():
    """
    Specifying a custom mapping of JSON key to dataclass field, via the
    `json_key` helper function.
    """

    @dataclass
    class MyClass(JSONSerializable):
        my_str: Annotated[str, json_key('myCustomStr', all=True)]
        my_bool: Annotated[bool, json_key(
            'my_json_bool', 'myTestBool', all=True)]

    value = 'Testing'
    expected = {'myCustomStr': value, 'my_json_bool': True}

    c = MyClass(my_str=value, my_bool=True)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    result = c.to_dict()
    log.debug('Parsed object: %r', result)

    assert result == expected


def test_to_dict_with_skip_defaults():
    """
    When `skip_defaults` is enabled in the class Meta, fields with default
    values should be excluded from the serialization process.
    """

    @dataclass
    class MyClass(JSONWizard):
        class _(JSONWizard.Meta):
            skip_defaults = True

        my_str: str
        other_str: str = 'any value'
        optional_str: str = None
        my_list: List[str] = field(default_factory=list)
        my_dict: DefaultDict[str, List[float]] = field(
            default_factory=lambda: defaultdict(list))

    c = MyClass('abc')
    log.debug('Instance: %r', c)

    out_dict = c.to_dict()
    assert out_dict == {'myStr': 'abc'}


def test_to_dict_with_excluded_fields():
    """
    Excluding dataclass fields from the serialization process works
    as expected.
    """

    @dataclass
    class MyClass(JSONWizard):

        my_str: str
        other_str: Annotated[str, json_key('AnotherStr', dump=False)]
        my_bool: bool = json_field('TestBool', dump=False)
        my_int: int = 3

    data = {'MyStr': 'my string',
            'AnotherStr': 'testing 123',
            'TestBool': True}

    c = MyClass.from_dict(data)
    log.debug('Instance: %r', c)

    # dynamically exclude the `my_int` field from serialization
    additional_exclude = ('my_int', )

    out_dict = c.to_dict(exclude=additional_exclude)
    assert out_dict == {'myStr': 'my string'}


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_set(input, expected, expectation):

    @dataclass
    class MyClass(JSONSerializable):
        num_set: Set[int]
        any_set: set

    # Sort expected so the assertions succeed
    expected = sorted(expected)

    input_set = set(input)
    c = MyClass(num_set=input_set, any_set=input_set)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('numSet', 'anySet'))

        # Set should be converted to list or tuple, as only those are JSON
        # serializable.
        assert isinstance(result['numSet'], (list, tuple))
        assert isinstance(result['anySet'], (list, tuple))

        assert sorted(result['numSet']) == expected
        assert sorted(result['anySet']) == expected


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_frozenset(input, expected, expectation):

    @dataclass
    class MyClass(JSONSerializable):
        num_set: FrozenSet[int]
        any_set: frozenset

    # Sort expected so the assertions succeed
    expected = sorted(expected)

    input_set = frozenset(input)
    c = MyClass(num_set=input_set, any_set=input_set)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('numSet', 'anySet'))

        # Set should be converted to list or tuple, as only those are JSON
        # serializable.
        assert isinstance(result['numSet'], (list, tuple))
        assert isinstance(result['anySet'], (list, tuple))

        assert sorted(result['numSet']) == expected
        assert sorted(result['anySet']) == expected


@pytest.mark.parametrize(
    'input,expected,expectation',
    [
        ({1, 2, 3}, [1, 2, 3], does_not_raise()),
        ((3.22, 2.11, 1.22), [3.22, 2.11, 1.22], does_not_raise()),
    ]
)
def test_deque(input, expected, expectation):

    @dataclass
    class MyQClass(JSONSerializable):
        num_deque: deque[int]
        any_deque: deque

    input_deque = deque(input)
    c = MyQClass(num_deque=input_deque, any_deque=input_deque)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)

        assert all(key in result for key in ('numDeque', 'anyDeque'))

        # Set should be converted to list or tuple, as only those are JSON
        # serializable.
        assert isinstance(result['numDeque'], list)
        assert isinstance(result['anyDeque'], list)

        assert result['numDeque'] == expected
        assert result['anyDeque'] == expected


@pytest.mark.parametrize(
    'input,expectation',
    [
        ('testing', pytest.raises(ParseError)),
        ('e1', does_not_raise()),
        (False, pytest.raises(ParseError)),
        (0, does_not_raise()),
    ]
)
@pytest.mark.xfail(reason='still need to add the dump hook for this type')
def test_literal(input, expectation):

    @dataclass
    class MyClass(JSONSerializable):
        class Meta(JSONSerializable.Meta):
            key_transform_with_dump = 'PASCAL'

        my_lit: Literal['e1', 'e2', 0]

    c = MyClass(my_lit=input)
    expected = {'MyLit': input}

    with expectation:
        actual = c.to_dict()

        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        (UUID('12345678-1234-1234-1234-1234567abcde'), does_not_raise()),
        (UUID('{12345678-1234-5678-1234-567812345678}'), does_not_raise()),
        (UUID('12345678123456781234567812345678'), does_not_raise()),
        (UUID('urn:uuid:12345678-1234-5678-1234-567812345678'), does_not_raise()),
    ]
)
def test_uuid(input, expectation):

    @dataclass
    class MyClass(JSONSerializable):
        class Meta(JSONSerializable.Meta):
            key_transform_with_dump = 'Snake'

        my_id: UUID

    c = MyClass(my_id=input)
    expected = {'my_id': input.hex}

    with expectation:
        actual = c.to_dict()

        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        (timedelta(seconds=12345), does_not_raise()),
        (timedelta(hours=1, minutes=32), does_not_raise()),
        (timedelta(days=1, minutes=51, seconds=7), does_not_raise()),
    ]
)
def test_timedelta(input, expectation):

    @dataclass
    class MyClass(JSONSerializable):
        class Meta(JSONSerializable.Meta):
            key_transform_with_dump = 'Snake'
        my_td: timedelta

    c = MyClass(my_td=input)
    expected = {'my_td': str(input)}

    with expectation:
        actual = c.to_dict()

        assert actual == expected
        log.debug('Parsed object: %r', actual)


@pytest.mark.parametrize(
    'input,expectation',
    [
        (
            {}, pytest.raises(ParseError)),
        (
            {'key': 'value'}, pytest.raises(ParseError)),
        (
            {'my_str': 'test', 'my_int': 2,
             'my_bool': True, 'other_key': 'testing'}, does_not_raise()),
        (
            {'my_str': 3}, pytest.raises(ParseError)),
        (
            {'my_str': 'test', 'my_int': 'test', 'my_bool': True},
            pytest.raises(ValueError)),
        (
            {'my_str': 'test', 'my_int': 2, 'my_bool': True},
            does_not_raise(),
        )
    ]
)
@pytest.mark.xfail(reason='still need to add the dump hook for this type')
def test_typed_dict(input, expectation):

    class MyDict(TypedDict):
        my_str: str
        my_bool: bool
        my_int: int

    @dataclass
    class MyClass(JSONSerializable):
        my_typed_dict: MyDict

    c = MyClass(my_typed_dict=input)

    with expectation:
        result = c.to_dict()
        log.debug('Parsed object: %r', result)


def test_using_dataclass_in_dict():
    """
    Using dataclass in a dictionary (i.e., dict[str, Test])
    works as expected.

    See https://github.com/rnag/dataclass-wizard/issues/159
    """
    @dataclass
    class Test:
        field: str

    @dataclass
    class Config:
        tests: dict[str, Test]

    config = {"tests": {"test_a": {"field": "a"}, "test_b": {"field": "b"}}}

    assert fromdict(Config, config) == Config(
        tests={'test_a': Test(field='a'),
               'test_b': Test(field='b')})


def test_bytes_and_bytes_array_are_supported():
    """Confirm dump with `bytes` and `bytesarray` is supported."""

    @dataclass
    class Foo(JSONWizard):
        b: bytes = None
        barray: bytearray = None
        s: str = None

    data = {'b': 'AAAA', 'barray': 'SGVsbG8sIFdvcmxkIQ==', 's': 'foobar'}

    # noinspection PyTypeChecker
    foo = Foo(b=b64decode('AAAA'),
              barray=bytearray(b'Hello, World!'),
              s='foobar')

    # noinspection PyTypeChecker
    assert foo.to_dict() == data



from dataclasses import MISSING
from typing import Any

from ..log import LOG


def is_builtin_class(cls: type) -> bool:
    """Check if a class is a builtin in Python."""
    return cls.__module__ == 'builtins'


class FunctionBuilder:
    __slots__ = (
        'current_function',
        'prev_function',
        'functions',
        'globals',
        'indent_level',
        'namespace',
    )

    def __init__(self):
        self.functions = {}
        self.indent_level = 0
        self.globals = {}
        self.namespace = {}

    def __ior__(self, other):
        """
        Allows `|=` operation for :class:`FunctionBuilder` objects,
        e.g. ::
            my_fn_builder |= other_fn_builder

        """
        self.functions |= other.functions
        self.globals |= other.globals
        return self

    def __enter__(self):
        self.indent_level += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        indent_lvl = self.indent_level = self.indent_level - 1

        if not indent_lvl:
            self.finalize_function()

    # noinspection PyAttributeOutsideInit
    def function(self, name: str, args: list, return_type=MISSING,
                 locals=None) -> 'FunctionBuilder':
        """Start a new function definition with optional return type."""
        curr_fn = getattr(self, 'current_function', None)
        if curr_fn is not None:
            curr_fn['indent_level'] = self.indent_level
            self.prev_function = curr_fn

        self.current_function = {
            "name": name,
            "args": args,
            "body": [],
            "return_type": return_type,
            "locals": locals if locals is not None else {},
        }

        self.indent_level = 0
        return self

    def _with_new_block(self,
                        name: str,
                        condition: 'str | None' = None,
                        comment: Any = '') -> 'FunctionBuilder':
        """Creates a new block. Used with a context manager (with)."""
        indent = '  ' * self.indent_level

        if comment:
            comment = f'  # {comment}'

        if condition is not None:
            self.current_function["body"].append(f"{indent}{name} {condition}:{comment}")
        else:
            self.current_function["body"].append(f"{indent}{name}:{comment}")

        return self

    def for_(self, condition: str) -> 'FunctionBuilder':
        """Equivalent to the `for` statement in Python.

        Sample Usage:

            >>> with FunctionBuilder().for_('i in range(3)'):
            >>>     ...

        Will generate the following code:

            >>> for i in range(3):
            >>>     ...

        """
        return self._with_new_block('for', condition)

    def if_(self, condition: str, comment: Any = '') -> 'FunctionBuilder':
        """Equivalent to the `if` statement in Python.

        Sample Usage:

            >>> with FunctionBuilder().if_('something is True'):
            >>>     ...

        Will generate the following code:

            >>> if something is True:
            >>>     ...

        """
        return self._with_new_block('if', condition, comment)

    def elif_(self, condition: str) -> 'FunctionBuilder':
        """Equivalent to the `elif` statement in Python.

        Sample Usage:

            >>> with FunctionBuilder().elif_('something is True'):
            >>>     ...

        Will generate the following code:

            >>> elif something is True:
            >>>     ...

        """
        return self._with_new_block('elif', condition)

    def else_(self) -> 'FunctionBuilder':
        """Equivalent to the `else` statement in Python.

        Sample Usage:

            >>> with FunctionBuilder().else_():
            >>>     ...

        Will generate the following code:

            >>> else:
            >>>     ...

        """
        return self._with_new_block('else')

    def try_(self) -> 'FunctionBuilder':
        """Equivalent to the `try` block in Python.

        Sample Usage:

            >>> with FunctionBuilder().try_():
            >>>     ...

        Will generate the following code:

            >>> try:
            >>>     ...

        """
        return self._with_new_block('try')

    def except_(self,
                cls: type[Exception],
                var_name: 'str | None' = None,
                *custom_classes: type[Exception]):
        """Equivalent to the `except` block in Python.

        Sample Usage:

            >>> with FunctionBuilder().except_(TypeError, 'exc'):
            >>>     ...

        Will generate the following code:

            >>> except TypeError as exc:
            >>>     ...

        """
        cls_name = cls.__name__
        statement = f'{cls_name} as {var_name}' if var_name else cls_name

        if not is_builtin_class(cls):
            if cls_name not in self.globals:
                # TODO
                # LOG.debug('Ensuring class in globals, cls=%s', cls_name)
                self.globals[cls_name] = cls

        if custom_classes:
            for cls in custom_classes:
                if not is_builtin_class(cls):
                    cls_name = cls.__name__
                    if cls_name not in self.globals:
                        # LOG.debug('Ensuring class in globals, cls=%s', cls_name)
                        self.globals[cls_name] = cls

        return self._with_new_block('except', statement)

    def except_multi(self, *classes: type[Exception]):
        """Equivalent to the `except` block in Python.

        Sample Usage:

            >>> with FunctionBuilder().except_multi(AttributeError, TypeError, ValueError):
            >>>     ...

        Will generate the following code:

            >>> except (AttributeError, TypeError, ValueError):
            >>>     ...

        """
        if len(classes) == 1:
            statement = classes[0].__name__
        else:
            class_names = ', '.join([cls.__name__ for cls in classes])
            statement = f'({class_names})'

        return self._with_new_block('except', statement)

    def break_(self):
        """Equivalent to the `break` statement in Python."""
        self.add_line('break')

    def add_line(self, line: str):
        """Add a line to the current function's body with proper indentation."""
        indent = '  ' * self.indent_level
        self.current_function["body"].append(f"{indent}{line}")

    def add_lines(self, *lines: str):
        """Add lines to the current function's body with proper indentation."""
        indent = '  ' * self.indent_level
        self.current_function["body"].extend(
            [f"{indent}{line}" for line in lines]
        )

    def increase_indent(self):  # pragma: no cover
        """Increase indentation level for nested code."""
        self.indent_level += 1

    def decrease_indent(self):  # pragma: no cover
        """Decrease indentation level."""
        if self.indent_level > 1:
            self.indent_level -= 1

    def finalize_function(self):
        """Finalize the function code and add to the list of functions."""
        # Add the function body and don't re-add the function definition
        curr_fn = self.current_function
        func_code = '\n'.join(curr_fn["body"])
        self.functions[curr_fn["name"]] = {
            "args": curr_fn["args"],
            "return_type": curr_fn["return_type"],
            "locals": curr_fn["locals"],
            "code": func_code
        }

        if (prev_fn := getattr(self, 'prev_function', None)) is not None:
            self.indent_level = prev_fn.pop('indent_level')
            self.current_function = prev_fn
            self.prev_function = None
        else:
            self.current_function  # Reset current function

    def create_functions(self, _globals=None):
        """Create functions by compiling the code."""
        # Note that we may mutate locals. Callers beware!
        # The only callers are internal to this module, so no
        # worries about external callers.

        # Compute the text of the entire function.
        # txt = f' def {name}({args}){return_annotation}:\n{body}'

        # Build the function code for all functions
        # Free variables in exec are resolved in the global namespace.
        # The global namespace we have is user-provided, so we can't modify it for
        # our purposes. So we put the things we need into locals and introduce a
        # scope to allow the function we're creating to close over them.

        fn_name_locals_and_code = []

        for name, func in self.functions.items():
            args = ','.join(func['args'])
            body = func['code']
            return_type = func['return_type']
            locals = func['locals']

            return_annotation = ''
            if return_type is not MISSING:
                locals[f'__dataclass_{name}_return_type__'] = return_type
                return_annotation = f'->__dataclass_{name}_return_type__'

            fn_name_locals_and_code.append(
                (name,
                 locals,
                 f'def {name}({args}){return_annotation}:\n{body}')
            )

        txt = '\n'.join([
            f"def __create_{name}_fn__({', '.join(locals.keys())}):\n"
            f" {code}\n"
            f" return {name}"
            for name, locals, code in fn_name_locals_and_code
        ])

        # Print the generated code for debugging
        # logging.debug(f"Generated function code:\n{all_func_code}")
        LOG.debug("Generated function code:\n%s", txt)

        ns = {}

        # TODO
        _globals = self.globals if _globals is None else _globals | self.globals

        LOG.debug("Globals before function compilation: %s", _globals)

        exec(txt, _globals, ns)

        # TODO do we need self.namespace?
        final_ns = self.namespace = {}

        # TODO: add function to dependent function `locals` rather than to `globals`

        for name, locals, _ in fn_name_locals_and_code:
            _globals[name] = final_ns[name] = ns[f'__create_{name}_fn__'](**locals)

        # final_ns = self.namespace = {
        #     name: ns[f'__create_{name}_fn__'](**locals)
        #     for name, locals, _ in fn_name_locals_and_code
        # }

        # Print namespace for debugging
        LOG.debug("Namespace after function compilation: %s", final_ns)

        return final_ns



import json
from dataclasses import MISSING, Field
from datetime import date, datetime, time
from typing import Generic, Mapping, NewType, Any, TypedDict

from .constants import PY310_OR_ABOVE, PY314_OR_ABOVE
from .decorators import cached_property
from .type_def import T, DT, PyNotRequired
# noinspection PyProtectedMember
from .utils.dataclass_compat import _create_fn
from .utils.object_path import split_object_path
from .utils.type_conv import as_datetime, as_time, as_date


# Define a simple type (alias) for the `CatchAll` field
#
# The `type` statement is introduced in Python 3.12
# Ref: https://docs.python.org/3.12/reference/simple_stmts.html#type
#
# TODO: uncomment following usage of `type` statement
#   once we drop support for Python 3.9 - 3.11
# if PY312_OR_ABOVE:
#     type CatchAll = Mapping
CatchAll = NewType('CatchAll', Mapping)
# A date, time, datetime sub type, or None.
# DT_OR_NONE = Optional[DT]


class Extras(TypedDict):
    """
    "Extra" config that can be used in the load / dump process.
    """
    config: PyNotRequired['META']
    cls: type
    cls_name: str
    fn_gen: 'FunctionBuilder'
    locals: dict[str, Any]
    pattern: PyNotRequired['PatternedDT']


# noinspection PyShadowingBuiltins
def json_key(*keys: str, all=False, dump=True):
    return JSON(*keys, all=all, dump=dump)


# noinspection PyPep8Naming,PyShadowingBuiltins
def KeyPath(keys, all=True, dump=True):
    if isinstance(keys, str):
        keys = split_object_path(keys)

    return JSON(*keys, all=all, dump=dump, path=True)


# noinspection PyShadowingBuiltins
def json_field(keys, *,
               all=False, dump=True,
               default=MISSING,
               default_factory=MISSING,
               init=True, repr=True,
               hash=None, compare=True, metadata=None):

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')

    return JSONField(keys, all, dump, default, default_factory, init, repr,
                     hash, compare, metadata)


env_field = json_field


class JSON:

    __slots__ = ('keys',
                 'all',
                 'dump',
                 'path')

    # noinspection PyShadowingBuiltins
    def __init__(self, *keys, all=False, dump=True, path=False):

        self.keys = (split_object_path(keys)
                     if path and isinstance(keys, str) else keys)
        self.all = all
        self.dump = dump
        self.path = path


class JSONField(Field):

    __slots__ = ('json', )

    # In Python 3.14, dataclasses adds a new parameter to the :class:`Field`
    # constructor: `doc`
    #
    # Ref: https://docs.python.org/3.14/library/dataclasses.html#dataclasses.field
    if PY314_OR_ABOVE:  # pragma: no cover
        # noinspection PyShadowingBuiltins
        def __init__(
            self,
            keys,
            all: bool,
            dump: bool,
            default,
            default_factory,
            init,
            repr,
            hash,
            compare,
            metadata,
            path: bool = False,
        ):

            super().__init__(
                default,
                default_factory,
                init,
                repr,
                hash,
                compare,
                metadata,
                False,
                None,
            )

            if isinstance(keys, str):
                keys = split_object_path(keys) if path else (keys,)
            elif keys is ...:
                keys = ()

            self.json = JSON(*keys, all=all, dump=dump, path=path)

    # In Python 3.10, dataclasses adds a new parameter to the :class:`Field`
    # constructor: `kw_only`
    #
    # Ref: https://docs.python.org/3.10/library/dataclasses.html#dataclasses.dataclass
    elif PY310_OR_ABOVE:  # pragma: no cover
        # noinspection PyShadowingBuiltins
        def __init__(self, keys, all: bool, dump: bool,
                     default, default_factory, init, repr, hash, compare,
                     metadata, path: bool = False):

            super().__init__(default, default_factory, init, repr, hash,
                             compare, metadata, False)

            if isinstance(keys, str):
                keys = split_object_path(keys) if path else (keys,)
            elif keys is ...:
                keys = ()

            self.json = JSON(*keys, all=all, dump=dump, path=path)

    else:  # pragma: no cover
        # noinspection PyArgumentList,PyShadowingBuiltins
        def __init__(self, keys, all: bool, dump: bool,
                     default, default_factory, init, repr, hash, compare,
                     metadata, path: bool = False):

            super().__init__(default, default_factory, init, repr, hash,
                             compare, metadata)

            if isinstance(keys, str):
                keys = split_object_path(keys) if path else (keys,)
            elif keys is ...:
                keys = ()

            self.json = JSON(*keys, all=all, dump=dump, path=path)


# noinspection PyPep8Naming
def Pattern(pattern):
    return PatternedDT(pattern)


class _PatternBase:
    __slots__ = ()

    def __class_getitem__(cls, pattern):
        return PatternedDT(pattern, cls.__base__)

    __getitem__ = __class_getitem__


class DatePattern(date, _PatternBase):
    __slots__ = ()


class TimePattern(time, _PatternBase):
    __slots__ = ()


class DateTimePattern(datetime, _PatternBase):
    __slots__ = ()


class PatternedDT(Generic[DT]):

    # `cls` is the date/time/datetime type or subclass.
    # `pattern` is the format string to pass in to `datetime.strptime`.
    __slots__ = ('cls',
                 'pattern')

    def __init__(self, pattern, cls = None):
        self.cls = cls
        self.pattern = pattern

    def get_transform_func(self):
        cls = self.cls

        # Parse with `fromisoformat` first, because its *much* faster than
        # `datetime.strptime` - see linked article above for more details.
        body_lines = [
            'dt = default_load_func(date_string, cls, raise_=False)',
            'if dt is not None:',
            '  return dt',
            'dt = datetime.strptime(date_string, pattern)',
        ]

        locals_ns = {'datetime': datetime,
                     'pattern': self.pattern,
                     'cls': cls}

        if cls is datetime:
            default_load_func = as_datetime
            body_lines.append('return dt')
        elif cls is date:
            default_load_func = as_date
            body_lines.append('return dt.date()')
        elif cls is time:
            default_load_func = as_time
            # temp fix for Python 3.11+, since `time.fromisoformat` is updated
            # to support more formats, such as "-" and "+" in strings.
            if '-' in self.pattern or '+' in self.pattern:
                body_lines = ['try:',
                              '  return datetime.strptime(date_string, pattern).time()',
                              'except (ValueError, TypeError):',
                              '  dt = default_load_func(date_string, cls, raise_=False)',
                              '  if dt is not None:',
                              '    return dt']
            else:
                body_lines.append('return dt.time()')
        elif issubclass(cls, datetime):
            default_load_func = as_datetime
            locals_ns['datetime'] = cls
            body_lines.append('return dt')
        elif issubclass(cls, date):
            default_load_func = as_date
            body_lines.append('return cls(dt.year, dt.month, dt.day)')
        elif issubclass(cls, time):
            default_load_func = as_time
            # temp fix for Python 3.11+, since `time.fromisoformat` is updated
            # to support more formats, such as "-" and "+" in strings.
            if '-' in self.pattern or '+' in self.pattern:
                body_lines = ['try:',
                              '  dt = datetime.strptime(date_string, pattern).time()',
                              'except (ValueError, TypeError):',
                              '  dt = default_load_func(date_string, cls, raise_=False)',
                              '  if dt is not None:',
                              '    return dt']

            body_lines.append('return cls(dt.hour, dt.minute, dt.second, '
                              'dt.microsecond, fold=dt.fold)')
        else:
            raise TypeError(f'Annotation for `Pattern` is of invalid type '
                            f'({cls}). Expected a type or subtype of: '
                            f'{DT.__constraints__}')

        locals_ns['default_load_func'] = default_load_func

        return _create_fn('pattern_to_dt',
                          ('date_string', ),
                          body_lines,
                          locals=locals_ns,
                          return_type=DT)

    def __repr__(self):
        repr_val = [f'{k}={getattr(self, k)!r}' for k in self.__slots__]
        return f'{self.__class__.__name__}({", ".join(repr_val)})'


class Container(list[T]):

    __slots__ = ('__dict__',
                 '__orig_class__')

    @cached_property
    def __model__(self):

        try:
            # noinspection PyUnresolvedReferences
            return self.__orig_class__.__args__[0]
        except AttributeError:
            cls_name = self.__class__.__qualname__
            msg = (f'A {cls_name} object needs to be instantiated with '
                   f'a generic type T.\n\n'
                   'Example:\n'
                   f'  my_list = {cls_name}[T](...)')

            raise TypeError(msg) from None

    def __str__(self):

        import pprint
        return pprint.pformat(self)

    def prettify(self, encoder = json.dumps,
                 ensure_ascii=False,
                 **encoder_kwargs):

        return self.to_json(
            indent=2,
            encoder=encoder,
            ensure_ascii=ensure_ascii,
            **encoder_kwargs
        )

    def to_json(self, encoder=json.dumps,
                **encoder_kwargs):

        from .loader_selection import asdict

        cls = self.__model__
        list_of_dict = [asdict(o, cls=cls) for o in self]

        return encoder(list_of_dict, **encoder_kwargs)

    def to_json_file(self, file, mode = 'w',
                     encoder=json.dump,
                     **encoder_kwargs):

        from .loader_selection import asdict

        cls = self.__model__
        list_of_dict = [asdict(o, cls=cls) for o in self]

        with open(file, mode) as out_file:
            encoder(list_of_dict, out_file, **encoder_kwargs)


# noinspection PyShadowingBuiltins
def path_field(keys, *,
               all=True, dump=True,
               default=MISSING,
               default_factory=MISSING,
               init=True, repr=True,
               hash=None, compare=True, metadata=None):

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')

    return JSONField(keys, all, dump, default, default_factory, init, repr,
                    hash, compare, metadata, True)


if PY314_OR_ABOVE:

    def skip_if_field(
        condition,
        *,
        default=MISSING,
        default_factory=MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=MISSING,
        doc=None,
    ):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError("cannot specify both default and default_factory")

        if metadata is None:
            metadata = {}

        metadata["__skip_if__"] = condition

        return Field(
            default, default_factory, init, repr, hash, compare, metadata, kw_only, doc
        )


# In Python 3.10, dataclasses adds a new parameter to the :class:`Field`
# constructor: `kw_only`
#
# Ref: https://docs.python.org/3.10/library/dataclasses.html#dataclasses.dataclass
elif PY310_OR_ABOVE:  # pragma: no cover
    def skip_if_field(condition, *, default=MISSING, default_factory=MISSING, init=True, repr=True,
                      hash=None, compare=True, metadata=None, kw_only=MISSING):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError('cannot specify both default and default_factory')

        if metadata is None:
            metadata = {}

        metadata['__skip_if__'] = condition

        return Field(default, default_factory, init, repr, hash,
                     compare, metadata, kw_only)
else:  # pragma: no cover
    def skip_if_field(condition, *, default=MISSING, default_factory=MISSING, init=True, repr=True,
                      hash=None, compare=True, metadata=None):

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError('cannot specify both default and default_factory')

        if metadata is None:
            metadata = {}

        metadata['__skip_if__'] = condition

        # noinspection PyArgumentList
        return Field(default, default_factory, init, repr, hash,
                     compare, metadata)


class Condition:

    __slots__ = (
        'op',
        'val',
        't_or_f',
        '_wrapped',
    )

    def __init__(self, operator, value):
        self.op = operator
        self.val = value
        self.t_or_f = operator in {'+', '!'}

    def __str__(self):
        return f"{self.op} {self.val!r}"

    def evaluate(self, other) -> bool:  # pragma: no cover
        # Optionally support runtime evaluation of the condition
        operators = {
            "==": lambda a, b: a == b,
            "!=": lambda a, b: a != b,
            "<": lambda a, b: a < b,
            "<=": lambda a, b: a <= b,
            ">": lambda a, b: a > b,
            ">=": lambda a, b: a >= b,
            "is": lambda a, b: a is b,
            "is not": lambda a, b: a is not b,
            "+": lambda a, _: True if a else False,
            "!": lambda a, _: not a,
        }
        return operators[self.op](other, self.val)


# Aliases for conditions

# noinspection PyPep8Naming
def EQ(value): return Condition("==", value)
# noinspection PyPep8Naming
def NE(value): return Condition("!=", value)
# noinspection PyPep8Naming
def LT(value): return Condition("<", value)
# noinspection PyPep8Naming
def LE(value): return Condition("<=", value)
# noinspection PyPep8Naming
def GT(value): return Condition(">", value)
# noinspection PyPep8Naming
def GE(value): return Condition(">=", value)
# noinspection PyPep8Naming
def IS(value): return Condition("is", value)
# noinspection PyPep8Naming
def IS_NOT(value): return Condition("is not", value)
# noinspection PyPep8Naming
def IS_TRUTHY(): return Condition("+", None)
# noinspection PyPep8Naming
def IS_FALSY(): return Condition("!", None)


# noinspection PyPep8Naming
def SkipIf(condition):
    """
    Mark a condition to be used as a skip directive during serialization.
    """
    condition._wrapped = True  # Set a marker attribute
    return condition


# Convenience alias, to skip serializing field if value is None
SkipIfNone = SkipIf(IS(None))


def finalize_skip_if(skip_if, operand_1, conditional):
    """
    Finalizes the skip condition by generating the appropriate string based on the condition.

    Args:
        skip_if (Condition): The condition to evaluate, containing truthiness and operation info.
        operand_1 (str): The primary operand for the condition (e.g., a variable or value).
        conditional (str): The conditional operator to use (e.g., '==', '!=').

    Returns:
        str: The resulting skip condition as a string.

    Example:
        >>> cond = Condition(t_or_f=True, op='+', val=None)
        >>> finalize_skip_if(cond, 'my_var', '==')
        'my_var'
    """
    if skip_if.t_or_f:
        return operand_1 if skip_if.op == '+' else f'not {operand_1}'

    return f'{operand_1} {conditional}'


def get_skip_if_condition(skip_if, _locals, operand_2=None, condition_i=None, condition_var='_skip_if_'):
    """
    Retrieves the skip condition based on the provided `Condition` object.

    Args:
        skip_if (Condition): The condition to evaluate.
        _locals (dict[str, Any]): A dictionary of local variables for condition evaluation.
        operand_2 (str): The secondary operand (e.g., a variable or value).
        condition_i (Condition): The condition var index.
        condition_var (str): The variable name to evaluate.

    Returns:
        Any: The result of the evaluated condition or a string representation for custom values.

    Example:
        >>> cond = Condition(t_or_f=False, op='==', val=10)
        >>> locals_dict = {}
        >>> get_skip_if_condition(cond, locals_dict, 'other_var')
        '== other_var'
    """
    # TODO: To avoid circular import
    from .class_helper import is_builtin

    if skip_if is None:
        return False

    if skip_if.t_or_f:  # Truthy or falsy condition, no operand
        return True

    if is_builtin(skip_if.val):
        return str(skip_if)

    # Update locals (as `val` is not a builtin)
    if operand_2 is None:
        operand_2 = f'{condition_var}{condition_i}'

    _locals[operand_2] = skip_if.val
    return f'{skip_if.op} {operand_2}'



from typing import overload
from weakref import WeakKeyDictionary

from ._type_def import _ENV_META, _META, META

META_BY_DATACLASS: WeakKeyDictionary[type, META] = WeakKeyDictionary()
BASE_META_CLS: type | None = None

def set_base_meta_cls(base_meta_cls: type) -> None: ...

@overload
def get_meta(cls: type) -> META: ...

@overload
def get_meta(cls: type, base_cls: type[_ENV_META]) -> type[_ENV_META]: ...

@overload
def get_meta(cls: type, base_cls: type[_META]) -> type[_META]:
    """
    Retrieves the Meta config for the :class:`AbstractJSONWizard` subclass.

    This config is set when the inner :class:`Meta` is sub-classed.
    """
    ...

def create_meta(cls: type, cls_name: str | None = None, **kwargs) -> META:
    """
    Sets the Meta config for the :class:`AbstractJSONWizard` subclass.

    WARNING: Only use if the Meta config is undefined,
      e.g. `get_meta` for the `cls` returns `base_cls`.

    """



from dataclasses import MISSING

from ..errors import ParseError


def safe_get(data, path, default=MISSING, raise_=True):
    current_data = data
    p = path  # to avoid "unbound local variable" warnings

    try:
        for p in path:
            current_data = current_data[p]

        return current_data

    # IndexError -
    #   raised when `data` is a `list`, and we access an index that is "out of bounds"
    # KeyError -
    #   raised when `data` is a `dict`, and we access a key that is not present
    # AttributeError -
    #   raised when `data` is an invalid type, such as a `None`
    except (IndexError, KeyError, AttributeError) as e:
        if raise_ and default is MISSING:
            raise _format_err(e, current_data, path, p) from None
        return default

    # TypeError -
    #   raised when `data` is a `list`, but we try to use it like a `dict`
    except TypeError:
        e = TypeError('Invalid path')
        raise _format_err(e, current_data, path, p, True) from None


def v1_safe_get(data, path, raise_):
    current_data = data

    try:
        for p in path:
            current_data = current_data[p]

        return current_data

    # IndexError -
    #   raised when `data` is a `list`, and we access an index that is "out of bounds"
    # KeyError -
    #   raised when `data` is a `dict`, and we access a key that is not present
    # AttributeError -
    #   raised when `data` is an invalid type, such as a `None`
    except (IndexError, KeyError, AttributeError) as e:
        if raise_:
            p = locals().get('p', path)  # to suppress "unbound local variable"
            raise _format_err(e, current_data, path, p, True) from None

        return MISSING

    # TypeError -
    #   raised when `data` is a `list`, but we try to use it like a `dict`
    except TypeError:
        e = TypeError('Invalid path')
        p = locals().get('p', path)  # to suppress "unbound local variable"
        raise _format_err(e, current_data, path, p, True) from None


def v1_env_safe_get(data, first_key, path, raise_):
    from ..._type_conv import as_collection

    current_data = data

    try:
        current_data = as_collection(current_data[first_key])

        for p in path:
            current_data = current_data[p]

        return current_data

    # IndexError -
    #   raised when `data` is a `list`, and we access an index that is "out of bounds"
    # KeyError -
    #   raised when `data` is a `dict`, and we access a key that is not present
    # AttributeError -
    #   raised when `data` is an invalid type, such as a `None`
    except (IndexError, KeyError, AttributeError) as e:
        if raise_:
            path = [first_key] + list(path)
            p = locals().get('p', path)  # to suppress "unbound local variable"
            raise _format_err(e, current_data, path, p, True) from None

        return MISSING

    # TypeError -
    #   raised when `data` is a `list`, but we try to use it like a `dict`
    except TypeError:
        e = TypeError('Invalid path')
        path = [first_key] + list(path)
        p = locals().get('p', path)  # to suppress "unbound local variable"
        raise _format_err(e, current_data, path, p, True) from None


def _format_err(e, current_data, path, current_path, invalid_path=False):
    return ParseError(
        e, current_data, dict if invalid_path else None, 'load',
        path=' => '.join(repr(p) for p in path),
        current_path=repr(current_path),
    )


# What values are considered "truthy" when converting to a boolean type.
# noinspection SpellCheckingInspection
_TRUTHY_VALUES = frozenset(("True", "true"))

# What values are considered "falsy" when converting to a boolean type.
# noinspection SpellCheckingInspection
_FALSY_VALUES = frozenset(("False", "false"))


# Valid starting separators in our custom "object path",
# for example `a.b[c].d.[-1]` has 5 start separators.
_START_SEP = frozenset(('.', '['))


def split_object_path(_input):
    res = []
    s = ""
    start_new = True
    in_literal = False

    parsed_string_literal = False

    in_braces = False

    escape_next_quote = False
    quote_char = None
    possible_number = False

    for c in _input:
        if c in _START_SEP:
            if in_literal:
                s += c
            else:
                if c == '.':
                    # A period within braces [xxx] OR within a string "xxx",
                    # should be captured.
                    if in_braces:
                        s += c
                        continue
                    in_braces = False
                else:
                    in_braces = True

                start_new = True
                if s:
                    if possible_number:
                        possible_number = False
                        try:
                            num = int(s)
                            res.append(num)
                        except ValueError:
                            try:
                                num = float(s)
                                res.append(num)
                            except ValueError:
                                res.append(s)
                    elif parsed_string_literal:
                        parsed_string_literal = False
                        res.append(s)
                    else:
                        if s in _TRUTHY_VALUES:
                            res.append(True)
                        elif s in _FALSY_VALUES:
                            res.append(False)
                        else:
                            res.append(s)

                    s = ""
        elif c == '\\' and in_literal:
            escape_next_quote = True
        elif escape_next_quote:
            if c != quote_char:
                # It was not an escape character after all!
                s += '\\'
            # Capture escaped character
            s += c
            escape_next_quote = False
        elif c == quote_char:
            in_literal = False
            quote_char = None
            parsed_string_literal = True
        elif c in {'"', "'"} and start_new:
            start_new = False
            in_literal = True
            quote_char = c
        elif (c in {'+', '-'} or c.isdigit()) and start_new:
            start_new = False
            possible_number = True
            s += c
        elif start_new:
            start_new = False
            s += c
        elif c == ']':
            if in_literal:
                s += c
            else:
                in_braces = False
        else:
            s += c

    if s:
        if possible_number:
            try:
                num = int(s)
                res.append(num)
            except ValueError:
                try:
                    num = float(s)
                    res.append(num)
                except ValueError:
                    res.append(s)
        elif parsed_string_literal:
            res.append(s)
        else:
            if s in _TRUTHY_VALUES:
                res.append(True)
            elif s in _FALSY_VALUES:
                res.append(False)
            else:
                res.append(s)

    return res



"""
Contains implementations for Abstract Base Classes
"""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, InitVar, Field
from typing import Type, TypeVar, Generic

from .models import Extras
from .type_def import T, TT


# Create a generic variable that can be 'AbstractJSONWizard', or any subclass.
W = TypeVar('W', bound='AbstractJSONWizard')


class AbstractEnvWizard(ABC):
    """
    Abstract class that defines the methods a sub-class must implement at a
    minimum to be considered a "true" Environment Wizard.
    """
    __slots__ = ()

    # Extends the `__annotations__` attribute to return only the fields
    # (variables) of the `EnvWizard` subclass.
    #
    # .. NOTE::
    #    This excludes fields marked as ``ClassVar``, or ones which are
    #    not type-annotated.
    __fields__: dict[str, Field]

    def dict(self):
        ...

    @abstractmethod
    def to_dict(self):
        ...

    @abstractmethod
    def to_json(self, indent=None):
        ...


class AbstractJSONWizard(ABC):

    __slots__ = ()

    @classmethod
    @abstractmethod
    def from_json(cls, string):
        ...

    @classmethod
    @abstractmethod
    def from_list(cls, o):
        ...

    @classmethod
    @abstractmethod
    def from_dict(cls, o):
        ...

    @abstractmethod
    def to_dict(self):
        ...

    @abstractmethod
    def to_json(self, *,
                encoder=json.dumps,
                indent=None,
                **encoder_kwargs):
        ...

    @classmethod
    @abstractmethod
    def list_to_json(cls,
                     instances,
                     encoder=json.dumps,
                     indent=None,
                     **encoder_kwargs):
        ...


@dataclass
class AbstractParser(ABC, Generic[T, TT]):

    __slots__ = ('base_type', )

    # Please see `abstractions.pyi` for documentation on each field.

    cls: InitVar[Type]
    extras: InitVar[Extras]
    base_type: type[T]

    def __contains__(self, item):
        return type(item) is self.base_type

    @abstractmethod
    def __call__(self, o) -> TT:
        ...


class AbstractLoader(ABC):

    __slots__ = ()

    @staticmethod
    @abstractmethod
    def transform_json_field(string):
        ...

    @staticmethod
    @abstractmethod
    def default_load_to(o, _):
        ...

    @staticmethod
    @abstractmethod
    def load_after_type_check(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_str(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_int(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_float(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_bool(o, _):
        ...

    @staticmethod
    @abstractmethod
    def load_to_enum(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_uuid(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_iterable(
            o, base_type,
            elem_parser):
        ...

    @staticmethod
    @abstractmethod
    def load_to_tuple(
            o, base_type,
            elem_parsers):
        ...

    @staticmethod
    @abstractmethod
    def load_to_named_tuple(
            o, base_type,
            field_to_parser,
            field_parsers):
        ...

    @staticmethod
    @abstractmethod
    def load_to_named_tuple_untyped(
            o, base_type,
            dict_parser, list_parser):
        ...

    @staticmethod
    @abstractmethod
    def load_to_dict(
            o, base_type,
            key_parser,
            val_parser):
        ...

    @staticmethod
    @abstractmethod
    def load_to_defaultdict(
            o, base_type,
            default_factory,
            key_parser,
            val_parser):
        ...

    @staticmethod
    @abstractmethod
    def load_to_typed_dict(
            o, base_type,
            key_to_parser,
            required_keys,
            optional_keys):
        ...

    @staticmethod
    @abstractmethod
    def load_to_decimal(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_datetime(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_time(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_date(o, base_type):
        ...

    @staticmethod
    @abstractmethod
    def load_to_timedelta(o, base_type):
        ...

    # @staticmethod
    # @abstractmethod
    # def load_func_for_dataclass(
    #     cls: Type[T],
    #     config: Optional[META],
    # ) -> Callable[[JSONObject], T]:
    #     """
    #     Generate and return the load function for a (nested) dataclass of
    #     type `cls`.
    #     """

    @classmethod
    @abstractmethod
    def get_parser_for_annotation(cls, ann_type,
                                  base_cls=None,
                                  extras=None):
        ...


class AbstractDumper(ABC):
    __slots__ = ()



import pytest


@pytest.fixture(scope='session')
def n():
    return 100_000


def pytest_addoption(parser):
    parser.addoption(
        "--all",  # long option
        "-A",
        action="store_true",
        default=False,
        help="Run benchmarks for *all* libraries, including *slower* ones like `jsons`",
    )



from collections.abc import Sequence
from typing import Any, TypeAlias

PathPart: TypeAlias = str | int | float | bool
PathType: TypeAlias = Sequence[PathPart]


def safe_get(data: dict | list,
             path: PathType,
             raise_: bool) -> Any:
    """
    Retrieve a value from a nested structure safely.

    Traverses a nested structure (e.g., dictionaries or lists) following a sequence of keys or indices specified in `path`.
    Handles missing keys, out-of-bounds indices, or invalid types gracefully.

    Args:
        data (Any): The nested structure to traverse.
        path (Iterable): A sequence of keys or indices to follow.
        raise_ (bool): True to raise an error on invalid path.

    Returns:
        Any: The value at the specified path, or `MISSING` if traversal fails.

    Raises:
        KeyError, IndexError, AttributeError, TypeError: If `default` is not provided
        and an error occurs during traversal.
    """
    ...


def env_safe_get(data: dict | list,
                 first_key: PathPart,
                 path: PathType,
                 raise_: bool) -> Any:
    """
    Retrieve a value from a nested structure safely.

    Traverses a nested structure (e.g., dictionaries or lists) following a sequence of keys or indices specified in `path`.
    Handles missing keys, out-of-bounds indices, or invalid types gracefully.

    Args:
        data (Any): The nested structure to traverse.
        first_key (Iterable): The first key in the path.
        path (Iterable): A sequence of keys or indices to follow.
        raise_ (bool): True to raise an error on invalid path.

    Returns:
        Any: The value at the specified path, or `MISSING` if traversal fails.

    Raises:
        KeyError, IndexError, AttributeError, TypeError: If `default` is not provided
        and an error occurs during traversal.
    """
    ...


def _format_err(e: Exception,
                current_data: Any,
                path: PathType,
                current_path: PathPart):
    """Format and return a `ParseError`."""
    ...


def split_object_path(_input: str) -> PathType:
    """
    Parse a custom object path string into a list of components.

    This function interprets a custom object path syntax and breaks it into individual path components,
    including dictionary keys, list indices, attributes, and nested elements.
    It handles escaped characters and supports mixed types (e.g., strings, integers, floats, booleans).

    Args:
        _input (str): The object path string to parse.

    Returns:
        PathType: A list of components representing the parsed path. Components can be strings,
                  integers, floats, booleans, or other valid key/index types.

    Example:
        >>> split_object_path(r'''a[b][c]["d\\\"o\\\""][e].f[go]['1'].then."y\\e\\\"s"[1]["we can!"].five.2.3.[ok][4.56].[-7.89].'let\\'sd\\othisy\\'all!'.yeah.123.False['True'].thanks!''')
        ['a', 'b', 'c', 'd"o"', 'e', 'f', 'go', '1', 'then', 'y\\e"s', 1, 'we can!', 'five', 2, 3, 'ok', 4.56, -7.89,
         "let'sd\\othisy'all!", 'yeah', 123, False, 'True', 'thanks!']
    """



import glob
import json
import os
import shutil
import subprocess
import matplotlib.pyplot as plt


def run_benchmarks():
    # Ensure the `.benchmarks` folder exists
    os.makedirs(".benchmarks", exist_ok=True)

    # Run pytest benchmarks and save results
    print("Running benchmarks...")
    result = subprocess.run(
        ["pytest", "benchmarks/catch_all.py", "--benchmark-save=benchmark_results"],
        capture_output=True,
        text=True
    )
    print(result.stdout)


def load_benchmark_results(file_path):
    """Load the benchmark results from the provided JSON file."""
    with open(file_path, "r") as f:
        return json.load(f)


def plot_relative_performance(results):
    """Plot relative performance for different benchmark groups."""
    benchmarks = results["benchmarks"]

    # Extract and format data
    names = []
    ops = []
    for bm in benchmarks:
        group = bm.get("group", "")
        library = "dataclass-wizard" if "wizard" in bm["name"] else "dataclasses-json"
        formatted_name = f"{group} ({library})"
        names.append(formatted_name)
        ops.append(bm["stats"]["ops"])

    # Calculate relative performance (ratio of each ops to the slowest ops)
    baseline = min(ops)
    relative_performance = [op / baseline for op in ops]

    # Plot bar chart
    plt.figure(figsize=(10, 6))
    bars = plt.barh(names, relative_performance, color="skyblue")
    plt.xlabel("Performance Relative to Slowest (times faster)")
    plt.title("Catch All: Relative Performance of dataclass-wizard vs dataclasses-json")
    plt.tight_layout()

    # Add data labels to the bars
    for bar, rel_perf in zip(bars, relative_performance):
        plt.text(bar.get_width() + 0.1, bar.get_y() + bar.get_height() / 2,
                 f"{rel_perf:.1f}x", va="center")

    # Save and display the plot
    plt.savefig("catch_all.png")
    plt.show()


def find_latest_benchmark_file():
    """Find the most recent benchmark result file."""
    benchmark_dir = ".benchmarks"
    pattern = os.path.join(benchmark_dir, "**", "*.json")
    files = glob.glob(pattern, recursive=True)
    if not files:
        raise FileNotFoundError("No benchmark files found.")
    latest_file = max(files, key=os.path.getctime)  # Find the most recently created file
    return latest_file


if __name__ == "__main__":
    # Step 1: Run benchmarks
    run_benchmarks()

    # Step 2: Find the latest benchmark results file
    benchmark_file = find_latest_benchmark_file()
    print(f"Latest benchmark file: {benchmark_file}")

    # Step 3: Load the benchmark results
    if os.path.exists(benchmark_file):
        results = load_benchmark_results(benchmark_file)

        # Step 4: Plot results
        plot_relative_performance(results)

    else:
        print(f"Benchmark file not found: {benchmark_file}")

    # Step 5: Move the generated image to docs folder for easy access
    shutil.copy("catch_all.png", "docs/")



import json

from .._decorators import cached_property
from .._type_def import Encoder, FileEncoder, T

class Container(list[T]):
    """Convenience wrapper around a collection of dataclass instances.

    For all intents and purposes, this should behave exactly as a `list`
    object.

    Usage:

        >>> from dataclass_wizard.utils.containers import Container
        >>> from dataclass_wizard import fromlist
        >>> from dataclasses import make_dataclass
        >>>
        >>> A = make_dataclass('A', [('f1', str), ('f2', int)])
        >>> list_of_a = fromlist(A, [{'f1': 'hello', 'f2': 1}, {'f1': 'world', 'f2': 2}])
        >>> c = Container[A](list_of_a)
        >>> print(c.prettify())

    """

    __slots__ = ('__dict__',
                 '__orig_class__')

    @cached_property
    def __model__(self) -> type[T]:
        """
        Given a declaration like Container[T], this returns the subscripted
        value of the generic type T.
        """
        ...

    def __init_subclass__(cls,
                          str=False):
        ...

    def prettify(self, encoder: Encoder = json.dumps,
                 indent=2,
                 ensure_ascii=False,
                 **encoder_kwargs) -> str:
        """
        Convert the list of instances to a *prettified* JSON string.
        """
        ...

    def to_json(self, encoder: Encoder = json.dumps,
                **encoder_kwargs) -> str:
        """
        Convert the list of instances to a JSON string.
        """
        ...

    def to_json_file(self, file: str, mode: str = 'w',
                     encoder: FileEncoder = json.dump,
                     **encoder_kwargs) -> None:
        """
        Serializes the list of instances and writes it to a JSON file.
        """
        ...



from __future__ import annotations

import datetime
import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

import pytest

from dataclass_wizard import JSONWizard, DumpMeta
from dataclass_wizard.errors import ParseError
from ..conftest import *

log = logging.getLogger(__name__)


@dataclass
class B:
    date_field: datetime.datetime | None


@dataclass
class C:
    ...


@dataclass
class D:
    ...


@dataclass
class DummyClass:
    ...


@pytest.mark.parametrize(
    'input,expectation',
    [
        # OK: `my_field1` is passed in a float (not in valid Union types); parses as str
        ({'my_field1': 3.1, 'my_field2': [], 'my_field3': (3,)}, does_not_raise()),
        # Wrong type: `my_field3` is passed a float type
        ({'my_field1': 3, 'my_field2': [], 'my_field3': 2.1}, pytest.raises(ParseError)),
        # OK: `my_field3` is passed a list type
        ({'my_field1': 3, 'my_field2': [], 'my_field3': [1]}, does_not_raise()),
        # OK: `my_field3` is passed in a tuple of float (parses as tuple of int)
        ({'my_field1': 3, 'my_field2': [], 'my_field3': (1.0,)}, does_not_raise()),
        # OK: `my_field3` is passed in a tuple of int (one of the valid Union types)
        ({'my_field1': 3, 'my_field2': [], 'my_field3': (1,)}, does_not_raise()),
        # Wrong number of elements for `my_field3`: expected only one
        ({'my_field1': 3, 'my_field2': [], 'my_field3': (1, 2)}, does_not_raise()),
        # Type checks for all fields
        ({'my_field1': 'string',
          'my_field2': [{'date_field': None}],
          'my_field3': ('hello world',)}, does_not_raise()),

    ]
)
def test_load_with_future_annotation_v1(input, expectation):
    """
    Test case using the latest Python 3.10 features, such as PEP 604- style
    annotations.

    Ref: https://www.python.org/dev/peps/pep-0604/
    """

    @dataclass
    class A(JSONWizard):
        my_field1: bool | str | int
        my_field2: list[B]
        my_field3: int | tuple[str | int]

    with expectation:
        result = A.from_dict(input)
        log.debug('Parsed object: %r', result)


@pytest.mark.parametrize(
    'input,expectation',
    [
        # technically wrong type: `my_field2` is passed in a float (expected str, int, or None)
        # but it parses as str
        ({'my_field1': datetime.date.min, 'my_field2': 1.23, 'my_field3': {'key': [None]}},
         does_not_raise()),
        # Type checks
        ({'my_field1': datetime.date.max, 'my_field2': None, 'my_field3': {'key': []}}, does_not_raise()),
        # ParseError: expected list of B, C, D, or None; passed in a list of string instead.
        ({'my_field1': Decimal('3.1'), 'my_field2': 7, 'my_field3': {'key': ['hello']}},
         does_not_raise()),
        # ParseError: expected list of B, C, D, or None; passed in a list of DummyClass instead.
        ({'my_field1': Decimal('3.1'), 'my_field2': 7, 'my_field3': {'key': [DummyClass()]}},
         does_not_raise()),
        # Type checks
        ({'my_field1': Decimal('3.1'), 'my_field2': 7, 'my_field3': {'key': [None]}},
         does_not_raise()),
        pytest.param({'my_field1': Decimal('3.1'), 'my_field2': 7, 'my_field3': {'key': [C()]}},
                     does_not_raise()),
    ]
)
def test_load_with_future_annotation_v2(input, expectation):
    """
    Test case using the latest Python 3.10 features, such as PEP 604- style
    annotations.

    Ref: https://www.python.org/dev/peps/pep-0604/
    """

    @dataclass
    class A(JSONWizard):
        class _(JSONWizard.Meta):
            unsafe_parse_dataclass_in_union = True

        my_field1: Decimal | datetime.date | str
        my_field2: str | Optional[int]
        my_field3: dict[str | int, list[B | C | Optional[D]]]

    with expectation:
        result = A.from_dict(input)
        log.debug('Parsed object: %r', result)


def test_dataclasses_in_union_types():
    """Dataclasses in Union types when manually specifying `tag` value."""

    @dataclass
    class Container(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'SNAKE'

        my_data: Data
        my_dict: dict[str, A | B]

    @dataclass
    class Data:
        my_str: str
        my_list: list[C | D]

    @dataclass
    class A(JSONWizard):
        class _(JSONWizard.Meta):
            tag = 'AA'

        val: str

    @dataclass
    class B(JSONWizard):
        class _(JSONWizard.Meta):
            tag = 'BB'

        val: int

    @dataclass
    class C(JSONWizard):
        class _(JSONWizard.Meta):
            tag = '_C_'

        my_field: int

    @dataclass
    class D(JSONWizard):
        class _(JSONWizard.Meta):
            tag = '_D_'

        my_field: float

    # Fix so the forward reference works
    globals().update(locals())

    c = Container.from_dict({
        'my_data': {
            'my_str': 'string',
            'my_list': [{'__tag__': '_D_', 'my_field': 1.23},
                        {'__tag__': '_C_', 'my_field': 3.0}]
        },
        'my_dict': {
            'key': {'__tag__': 'AA',
                    'val': '123'}
        }
    })

    expected_obj = Container(
        my_data=Data(my_str='string',
                     my_list=[D(my_field=1.23),
                              C(my_field=3)]),
        my_dict={'key': A(val='123')}
    )

    expected_dict = {
        "my_data": {"my_str": "string",
                    "my_list": [{"my_field": 1.23, "__tag__": "_D_"},
                                {"my_field": 3, "__tag__": "_C_"}]},
        "my_dict": {"key": {"val": "123", "__tag__": "AA"}}
    }

    assert c == expected_obj
    assert c.to_dict() == expected_dict


def test_dataclasses_in_union_types_with_auto_assign_tags():
    """
    Dataclasses in Union types with auto-assign tags, and a custom tag field.
    """
    @dataclass
    class Container(JSONWizard):
        class _(JSONWizard.Meta):
            dump_case = 'SNAKE'
            tag_key = 'type'
            auto_assign_tags = True

        my_data: Data
        my_dict: dict[str, A | B]

    @dataclass
    class Data:
        my_str: str
        my_list: list[C | D | E]

    @dataclass
    class A:
        val: str

    @dataclass
    class B:
        val: int

    @dataclass
    class C:
        my_field: int

    @dataclass
    class D:
        my_field: float

    @dataclass
    class E:
        ...

    # This is to coverage a case where we have a Meta config for a class,
    # but we do not define a tag in the Meta config.
    DumpMeta(key_transform='SNAKE').bind_to(D)

    # Bind a custom tag to class E, so we can cover a case when
    # `auto_assign_tags` is true, but we are still able to specify a
    # custom tag for a class.
    DumpMeta(tag='!E').bind_to(E)

    # Fix so the forward reference works
    globals().update(locals())

    c = Container.from_dict({
        'my_data': {
            'my_str': 'string',
            'my_list': [{'type': 'D', 'my_field': 1.23},
                        {'type': 'C', 'my_field': 3.0},
                        {'type': '!E'}]
        },
        'my_dict': {
            'key': {'type': 'A',
                    'val': '123'}
        }
    })

    expected_obj = Container(
        my_data=Data(my_str='string',
                     my_list=[D(my_field=1.23),
                              C(my_field=3),
                              E()]),
        my_dict={'key': A(val='123')}
    )

    expected_dict = {
        "my_data": {"my_str": "string",
                    "my_list": [{"my_field": 1.23, "type": "D"},
                                {"my_field": 3, "type": "C"},
                                {'type': '!E'}]},
        "my_dict": {"key": {"val": "123", "type": "A"}}
    }

    assert c == expected_obj
    assert c.to_dict() == expected_dict



"""
Contains implementations for Abstract Base Classes
"""
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, InitVar, Field
from datetime import datetime, time, date, timedelta
from decimal import Decimal
from typing import (
    Any, TypeVar, SupportsFloat, AnyStr,
    Text, Sequence, Iterable, Generic
)

from .models import Extras
from .type_def import (
    DefFactory, FrozenKeys, ListOfJSONObject, JSONObject, Encoder,
    M, N, T, TT, NT, E, U, DD, LSQ
)


# Create a generic variable that can be 'AbstractEnvWizard', or any subclass.
E = TypeVar('E', bound='AbstractEnvWizard')

# Create a generic variable that can be 'AbstractJSONWizard', or any subclass.
W = TypeVar('W', bound='AbstractJSONWizard')

FieldToParser = dict[str, AbstractParser]


class AbstractEnvWizard(ABC):
    """
    Abstract class that defines the methods a sub-class must implement at a
    minimum to be considered a "true" Environment Wizard.
    """
    __slots__ = ()

    # Extends the `__annotations__` attribute to return only the fields
    # (variables) of the `EnvWizard` subclass.
    #
    # .. NOTE::
    #    This excludes fields marked as ``ClassVar``, or ones which are
    #    not type-annotated.
    __fields__: dict[str, Field]

    def dict(self: E) -> JSONObject:
        """
        Same as ``__dict__``, but only returns values for fields defined
        on the `EnvWizard` instance. See :attr:`__fields__` for more info.

        .. NOTE::
           The values in the returned dictionary object are not needed to be
           JSON serializable. Use :meth:`to_dict` if this is required.
        """

    @abstractmethod
    def to_dict(self: E) -> JSONObject:
        """
        Converts an instance of a `EnvWizard` subclass to a Python dictionary
        object that is JSON serializable.
        """

    @abstractmethod
    def to_json(self: E, indent=None) -> AnyStr:
        """
        Converts an instance of a `EnvWizard` subclass to a JSON `string`
        representation.
        """


class AbstractJSONWizard(ABC):
    """
    Abstract class that defines the methods a sub-class must implement at a
    minimum to be considered a "true" JSON Wizard.

    In particular, these are the abstract methods which - if correctly
    implemented - will allow a concrete sub-class (ideally a dataclass) to
    be properly loaded from, and serialized to, JSON.

    """
    __slots__ = ()

    @classmethod
    @abstractmethod
    def from_json(cls: type[W], string: AnyStr) -> W | list[W]:
        """
        Converts a JSON `string` to an instance of the dataclass, or a list of
        the dataclass instances.
        """

    @classmethod
    @abstractmethod
    def from_list(cls: type[W], o: ListOfJSONObject) -> list[W]:
        """
        Converts a Python `list` object to a list of the dataclass instances.
        """

    @classmethod
    @abstractmethod
    def from_dict(cls: type[W], o: JSONObject) -> W:
        """
        Converts a Python `dict` object to an instance of the dataclass.
        """

    @abstractmethod
    def to_dict(self: W) -> JSONObject:
        """
        Converts the dataclass instance to a Python dictionary object that is
        JSON serializable.
        """

    @abstractmethod
    def to_json(self: W, *,
                encoder: Encoder = json.dumps,
                indent=None,
                **encoder_kwargs) -> str:
        """
        Converts the dataclass instance to a JSON `string` representation.
        """

    @classmethod
    @abstractmethod
    def list_to_json(cls: type[W],
                     instances: list[W],
                     encoder: Encoder = json.dumps,
                     **encoder_kwargs) -> str:
        """
        Converts a ``list`` of dataclass instances to a JSON `string`
        representation.
        """


@dataclass
class AbstractParser(ABC, Generic[T, TT]):
    """
    Abstract parsers, which will ideally act as dispatchers to route objects
    to the `load` or `dump` hook methods responsible for transforming the
    objects into the annotated type for the dataclass field for which value we
    want to set. The error handling logic should ideally be implemented on the
    Parser (dispatcher) side.

    There can be more complex Parsers, for example ones which will handle
    ``typing.Union``, ``typing.Literal``, ``Dict``, and ``NamedTuple`` types.
    There can even be nested Parsers, which will be useful for handling
    collection and sequence types.

    """
    __slots__ = ('base_type', )

    # This represents the class that contains the field that has an annotated
    # type `base_type`. This is primarily useful for resolving `ForwardRef`
    # types, where we need the globals of the class to resolve the underlying
    # type of the reference.
    cls: InitVar[type]

    # This represents an optional Meta config that was specified for the main
    # dataclass. This is primarily useful to have so that we can merge this
    # base Meta config with the one for each class, and then recursively
    # apply the merged Meta config to any nested dataclasses.
    extras: InitVar[Extras]

    # This is usually the underlying base type of the annotation (for example,
    # for `list[str]` it will be `list`), though in some cases this will be
    # the annotation itself.
    base_type: type[T]

    def __contains__(self, item) -> bool:
        """
        Return true if the Parser is expected to handle the specified item
        type. Checks against the exact type instead of `isinstance` so we can
        handle special cases like `bool`, which is a subclass of `int`.
        """

    @abstractmethod
    def __call__(self, o: Any) -> TT:
        """
        Parse object `o`
        """


class AbstractLoader(ABC):
    """
    Abstract loader which defines the helper methods that can be used to load
    an object `o` into an object of annotated (or concrete) type `base_type`.

    """
    __slots__ = ()

    @staticmethod
    @abstractmethod
    def transform_json_field(string: str) -> str:
        """
        Transform a JSON field name (which will typically be camel-cased) into
        the conventional format for a dataclass field name (which will ideally
        be snake-cased).
        """

    @staticmethod
    @abstractmethod
    def default_load_to(o: T, _: Any) -> T:
        """
        Default load function if no other paths match. Generally, this will
        be a stub load method.
        """

    @staticmethod
    @abstractmethod
    def load_after_type_check(o: Any, base_type: type[T]) -> T:
        """
        Load an object `o`, after confirming that it is indeed of
        type `base_type`.

        :raises ParseError: If the object is not of the expected type.
        """

    @staticmethod
    @abstractmethod
    def load_to_str(o: Text | N | None, base_type: type[str]) -> str:
        """
        Load a string or numeric type into a new object of type `base_type`
        (generally a sub-class of the :class:`str` type)
        """

    @staticmethod
    @abstractmethod
    def load_to_int(o: str | int | bool | None, base_type: type[N]) -> N:
        """
        Load a string or int into a new object of type `base_type`
        (generally a sub-class of the :class:`int` type)
        """

    @staticmethod
    @abstractmethod
    def load_to_float(o: SupportsFloat | str, base_type: type[N]) -> N:
        """
        Load a string or float into a new object of type `base_type`
        (generally a sub-class of the :class:`float` type)
        """

    @staticmethod
    @abstractmethod
    def load_to_bool(o: str | bool | N, _: type[bool]) -> bool:
        """
        Load a bool, string, or an numeric value into a new object of type
        `bool`.

        *Note*: `bool` cannot be sub-classed, so the `base_type` argument is
        discarded in this case.
        """

    @staticmethod
    @abstractmethod
    def load_to_enum(o: AnyStr | N, base_type: type[E]) -> E:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        sub-class of the :class:`Enum` type)
        """

    @staticmethod
    @abstractmethod
    def load_to_uuid(o: AnyStr | U, base_type: type[U]) -> U:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        sub-class of the :class:`UUID` type)
        """

    @staticmethod
    @abstractmethod
    def load_to_iterable(
            o: Iterable, base_type: type[LSQ],
            elem_parser: AbstractParser) -> LSQ:
        """
        Load a list, set, frozenset or deque into a new object of type
        `base_type` (generally a list, set, frozenset, deque, or a sub-class
        of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_tuple(
            o: list | tuple, base_type: type[tuple],
            elem_parsers: Sequence[AbstractParser]) -> tuple:
        """
        Load a list or tuple into a new object of type `base_type` (generally
        a :class:`tuple` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_named_tuple(
            o: dict | list | tuple, base_type: type[NT],
            field_to_parser: FieldToParser,
            field_parsers: list[AbstractParser]) -> NT:
        """
        Load a dictionary, list, or tuple to a `NamedTuple` sub-class
        """

    @staticmethod
    @abstractmethod
    def load_to_named_tuple_untyped(
            o: dict | list | tuple, base_type: type[NT],
            dict_parser: AbstractParser, list_parser: AbstractParser) -> NT:
        """
        Load a dictionary, list, or tuple to a (generally) un-typed
        `collections.namedtuple`
        """

    @staticmethod
    @abstractmethod
    def load_to_dict(
            o: dict, base_type: type[M],
            key_parser: AbstractParser,
            val_parser: AbstractParser) -> M:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        :class:`dict` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_defaultdict(
            o: dict, base_type: type[DD],
            default_factory: DefFactory,
            key_parser: AbstractParser,
            val_parser: AbstractParser) -> DD:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        :class:`collections.defaultdict` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_typed_dict(
            o: dict, base_type: type[M],
            key_to_parser: FieldToParser,
            required_keys: FrozenKeys,
            optional_keys: FrozenKeys) -> M:
        """
        Load an object `o` annotated as a ``TypedDict`` sub-class into a new
        object of type `base_type` (generally a :class:`dict` or a sub-class
        of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_decimal(o: N, base_type: type[Decimal]) -> Decimal:
        """
        Load an object `o` into a new object of type `base_type` (generally a
        :class:`Decimal` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_datetime(
            o: str | N, base_type: type[datetime]) -> datetime:
        """
        Load a string or number (int or float) into a new object of type
        `base_type` (generally a :class:`datetime` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_time(o: str, base_type: type[time]) -> time:
        """
        Load a string or number (int or float) into a new object of type
        `base_type` (generally a :class:`time` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_date(o: str | N, base_type: type[date]) -> date:
        """
        Load a string or number (int or float) into a new object of type
        `base_type` (generally a :class:`date` or a sub-class of one)
        """

    @staticmethod
    @abstractmethod
    def load_to_timedelta(
            o: str | N, base_type: type[timedelta]) -> timedelta:
        """
        Load a string or number (int or float) into a new object of type
        `base_type` (generally a :class:`timedelta` or a sub-class of one)
        """

    @classmethod
    @abstractmethod
    def get_parser_for_annotation(cls, ann_type: type[T],
                                  base_cls: type = None,
                                  extras: Extras = None) -> AbstractParser:
        """
        Returns the Parser (dispatcher) for a given annotation type.

        `base_cls` is the original class object, this is useful when the
        annotated type is a :class:`typing.ForwardRef` object
        """


class AbstractDumper(ABC):
    __slots__ = ()

    def __pre_as_dict__(self):
        """
        Optional hook that runs before the dataclass instance is processed and
        before it is converted to a dictionary object via :meth:`to_dict`.

        To override this, subclasses need to extend from :class:`DumpMixIn`
        and implement this method. A simple example is shown below:

        >>> from dataclasses import dataclass
        >>> from dataclass_wizard import JSONSerializable, DumpMixin
        >>>
        >>>
        >>> @dataclass
        >>> class MyClass(JSONSerializable, DumpMixin):
        >>>     my_str: str
        >>>
        >>>     def __pre_as_dict__(self):
        >>>         self.my_str = self.my_str.swapcase()

        @deprecated since v0.28.0. Use `_pre_dict()` instead - no need
            to subclass from DumpMixin.
        """
        ...



from dataclasses import dataclass, is_dataclass
from dataclass_wizard.v0 import JSONWizard


def test_jsonwizard_is_not_a_dataclass_mixin():
    # If JSONWizard becomes a dataclass again, frozen subclasses can break.
    assert not is_dataclass(JSONWizard)


def test_frozen_dataclass_can_inherit_from_jsonwizard():
    @dataclass(eq=False, frozen=True)
    class BaseClass(JSONWizard):
        x: int = 1

    obj = BaseClass()
    assert obj.x == 1



import sys

# Package name
PACKAGE_NAME: str
# Library Log Level
LOG_LEVEL: str
# Current system Python version
_PY_VERSION: tuple[int, int] = sys.version_info[:2]
# Check if currently running Python 3.x or higher
PY310_OR_ABOVE: bool
PY311_OR_ABOVE: bool
PY312_OR_ABOVE: bool
PY313_OR_ABOVE: bool
PY314_OR_ABOVE: bool
# The name of the dictionary object that contains `dump` or `load` hooks
_HOOKS: str
# Attribute names (mostly internal)
SINGLE_ARG_ALIAS: str
IDENTITY: str
TAG: str
CATCH_ALL: str



import logging
from textwrap import dedent
from unittest.mock import ANY

import pytest
from pytest_mock import MockerFixture

from dataclass_wizard.v0.wizard_cli import main, PyCodeGenerator
from ...conftest import data_file_path


log = logging.getLogger(__name__)


def gen_schema(filename: str):
    """
    Helper function to call `wiz gen-schema` and pass the full path to a test
    file in the `testdata` directory.
    """

    main(['gs', data_file_path(filename), '-'])


def assert_py_code(expected, capfd=None, py_code=None):
    """
    Helper function to assert that generated Python code is as expected.
    """
    if py_code is None:
        py_code = _get_captured_py_code(capfd)

    # TODO update to `info` level to see the output in terminal.
    log.debug('Generated Python code:\n%s\n%s',
              '-' * 20, py_code)

    assert py_code == dedent(expected).lstrip()


def _get_captured_py_code(capfd) -> str:
    """Reads the Python code which is written to stdout."""
    out, err = capfd.readouterr()
    assert not err

    py_code_lines = out.split('\n')[4:]
    py_code = '\n'.join(py_code_lines)

    return py_code


@pytest.fixture
def mock_path(mocker: MockerFixture):
    return mocker.patch('dataclass_wizard.v0.wizard_cli.schema.Path')


@pytest.fixture
def mock_stdin(mocker: MockerFixture):
    return mocker.patch('sys.stdin')


@pytest.fixture
def mock_open(mocker: MockerFixture):
    return mocker.patch('dataclass_wizard.v0.wizard_cli.cli.open')


def test_call_py_code_generator_with_file_name(mock_path):
    """
    Test calling the constructor for :class:`PyCodeGenerator` with the
    `file_name` argument. Added for code coverage.
    """
    mock_path().read_bytes.return_value = b'{"key": "1.23", "secondKey": null}'

    expected = '''
    from dataclasses import dataclass
    from typing import Any

    from dataclass_wizard import JSONWizard


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        key: float
        second_key: Any
    '''

    code_gen = PyCodeGenerator(file_name='my_file.txt',
                               force_strings=True)

    assert_py_code(expected, py_code=code_gen.py_code)


def test_call_py_code_generator_with_experimental_features():
    """
    Test calling the constructor for :class:`PyCodeGenerator` with the
    `-x|--experimental` flag.
    """

    string = """\
    {"someField": null, "Some_List": [],
     "Objects": [{"key1": false},
                 {"key1": 1.2, "key2": "string"},
                 {"key1": "val", "key2": null}]
    }\
    """

    expected = '''
    from __future__ import annotations

    from dataclasses import dataclass
    from typing import Any

    from dataclass_wizard import JSONWizard


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        some_field: Any
        some_list: list
        objects: list[Object]


    @dataclass
    class Object:
        """
        Object dataclass

        """
        key1: bool | float | str
        key2: str | None
    '''

    code_gen = PyCodeGenerator(file_contents=string,
                               experimental=True,
                               force_strings=True)

    assert_py_code(expected, py_code=code_gen.py_code)


def test_call_wiz_cli_without_subcommand():
    """
    Calling wiz-cli without a sub-command. Added for code coverage.
    """
    with pytest.raises(SystemExit) as e:
        main([])

    assert e.value.code == 0


def test_call_wiz_cli_with_invalid_json_input(capsys, mock_stdin):
    """
    Calling wiz-cli with invalid JSON as input. Added for code coverage.
    """
    invalid_json = '{"key": "value"'

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = invalid_json

    with capsys.disabled():
        with pytest.raises(SystemExit) as e:
            main(['gs', '-', '-'])

        assert 'JSONDecodeError' in e.value.code


def test_call_wiz_cli_with_invalid_json_type(capsys, mock_stdin):
    """
    Calling wiz-cli when input is valid JSON, but not a valid JSON object
    (list or dictionary type). Added for code coverage.
    """
    invalid_json = '"my string value"'

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = invalid_json

    with capsys.disabled():
        with pytest.raises(SystemExit) as e:
            main(['gs', '-', '-'])

        assert 'TypeError' in e.value.code


def test_call_wiz_cli_when_double_quotes_are_used_to_wrap_input(
        capsys, mock_stdin):
    """
    Calling wiz-cli when input is piped via stdin and the string is wrapped
    with double quotes instead of single quotes. Added for code coverage.
    """

    # Note: this can be the result of the following command:
    #   echo "{"key": "value"}" | wiz gs
    invalid_json = '\"{"key": "value"}\"'

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = invalid_json

    with capsys.disabled():
        with pytest.raises(SystemExit) as e:
            main(['gs', '-'])

        log.debug(e.value.code)
        assert 'double quotes' in e.value.code


def test_call_wiz_cli_with_mock_stdout(capsys, mock_stdin, mocker):
    """
    Calling wiz-cli with mock stdout. Added for code coverage.
    """
    valid_json = '{"key": "value"}'

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = valid_json

    with capsys.disabled():
        mock_stdout = mocker.patch('sys.stdout')
        mock_stdout.name = '<stdout>'
        mock_stdout.isatty.return_value = False

        main(['gs', '-', '-'])

    mock_stdout.write.assert_called()


def test_call_wiz_cli_with_output_filename_without_ext(
        mocker, mock_stdin, mock_open):
    """
    Calling wiz-cli with an output filename without an extension. The
    extension should automatically be added.
    """
    valid_json = '{"key": "value"}'

    mock_out = mocker.Mock()
    mock_out.name = 'testing'
    mock_out.fileno.return_value = 0

    mock_open.return_value = mock_out

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = valid_json

    main(['gs', '-', 'testing'])

    mock_open.assert_called_once_with(
        'testing.py', 'w', ANY, ANY, ANY)

    mock_out.write.assert_called_once()


def test_call_wiz_cli_when_open_raises_error(
        mocker, mock_stdin, mock_open):
    """
    Calling wiz-cli with an error is raised opening the JSON file.
    """
    valid_json = '{"key": "value"}'

    mock_open.side_effect = OSError

    mock_stdin.name = '<stdin>'
    mock_stdin.read.return_value = valid_json

    with pytest.raises(SystemExit) as e:
        main(['gs', '-', 'testing'])

    mock_open.assert_called_once()


def test_star_wars(capfd):

    expected = '''
    from dataclasses import dataclass
    from datetime import datetime
    from typing import List, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        name: str
        rotation_period: Union[int, str]
        orbital_period: Union[int, str]
        diameter: Union[int, str]
        climate: str
        gravity: str
        terrain: str
        surface_water: Union[int, str]
        population: Union[int, str]
        residents: List
        films: List[str]
        created: datetime
        edited: datetime
        url: str
    '''

    gen_schema('star_wars.json')

    assert_py_code(expected, capfd)


def test_input_1(capfd):

    expected = '''
    from dataclasses import dataclass

    from dataclass_wizard import JSONWizard


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        key: str
        int_key: int
        float_key: float
        my_dict: 'MyDict'


    @dataclass
    class MyDict:
        """
        MyDict dataclass

        """
        key2: str
    '''

    gen_schema('test1.json')

    assert_py_code(expected, capfd)


def test_input_2(capfd):

    expected = '''
    from dataclasses import dataclass
    from datetime import datetime
    from typing import Optional, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'
        field_1: int
        field_2: str


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        key: Optional[str]
        another_key: Optional[Union[str, int]]
        truth: int
        my_list: 'MyList'
        my_date: datetime
        my_id: str


    @dataclass
    class MyList:
        """
        MyList dataclass

        """
        pass
    '''

    gen_schema('test2.json')

    assert_py_code(expected, capfd)


def test_input_3(capfd):

    expected = '''
    from dataclasses import dataclass
    from typing import List, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'
        field_1: int
        field_2: int
        field_3: str
        field_4: bool


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        true_story: Union[str, int]
        true_bool: bool
        my_list: List[Union[int, 'MyList']]


    @dataclass
    class MyList:
        """
        MyList dataclass

        """
        hey: str
    '''

    gen_schema('test3.json')

    assert_py_code(expected, capfd)


def test_input_4(capfd):

    expected = '''
    from dataclasses import dataclass
    from typing import Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        input_index: int
        candidate_index: int
        delivery_line_1: str
        last_line: str
        delivery_point_barcode: Union[int, str]
        components: 'Components'
        metadata: 'Metadata'
        analysis: 'Analysis'


    @dataclass
    class Components:
        """
        Components dataclass

        """
        primary_number: Union[int, str]
        street_predirection: Union[bool, str]
        street_name: str
        street_suffix: str
        city_name: str
        state_abbreviation: str
        zipcode: Union[int, str]
        plus4_code: Union[int, str]
        delivery_point: Union[int, str]
        delivery_point_check_digit: Union[int, str]


    @dataclass
    class Metadata:
        """
        Metadata dataclass

        """
        record_type: str
        zip_type: str
        county_fips: Union[int, str]
        county_name: str
        carrier_route: str
        congressional_district: Union[int, str]
        rdi: str
        elot_sequence: Union[int, str]
        elot_sort: str
        latitude: float
        longitude: float
        precision: str
        time_zone: str
        utc_offset: int
        dst: bool


    @dataclass
    class Analysis:
        """
        Analysis dataclass

        """
        dpv_match_code: Union[bool, str]
        dpv_footnotes: str
        dpv_cmra: Union[bool, str]
        dpv_vacant: Union[bool, str]
        active: Union[bool, str]
    '''

    gen_schema('test4.json')

    assert_py_code(expected, capfd)


def test_input_5(capfd):

    expected = '''
    from dataclasses import dataclass
    from typing import List, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'
        field_1: List[Union[List[Union[str, 'Data2']], int, str]]


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        key: str


    @dataclass
    class Data2:
        """
        Data2 dataclass

        """
        key: int
        nested_classes: 'NestedClasses'


    @dataclass
    class NestedClasses:
        """
        NestedClasses dataclass

        """
        blah: str
        another_one: List['AnotherOne']
        just_something_with_a_space: int


    @dataclass
    class AnotherOne:
        """
        AnotherOne dataclass

        """
        testing: str
    '''

    gen_schema('test5.json')

    assert_py_code(expected, capfd)


def test_input_6(capfd):

    expected = '''
    from dataclasses import dataclass
    from datetime import date, time
    from typing import List, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        my_field: str
        another_field: date
        my_list: List[Union[int, 'MyList', List['Data2']]]


    @dataclass
    class MyList:
        """
        MyList dataclass

        """
        another_key: str


    @dataclass
    class Data2:
        """
        Data2 dataclass

        """
        key: str
        my_time: time
    '''

    gen_schema('test6.json')

    assert_py_code(expected, capfd)


def test_input_7(capfd):

    expected = '''
    from dataclasses import dataclass
    from typing import List, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        my_test_apis: List['MyTestApi']
        people: List['Person']
        children: List['Child']
        activities: List['Activity']
        equipment: List['Equipment']
        key: int
        nested_classes: 'NestedClasses'
        something_else: str


    @dataclass
    class MyTestApi:
        """
        MyTestApi dataclass

        """
        first_api: str


    @dataclass
    class Person:
        """
        Person dataclass

        """
        name: str
        age: Union[int, str]


    @dataclass
    class Child:
        """
        Child dataclass

        """
        name: str
        age: Union[int, float]


    @dataclass
    class Activity:
        """
        Activity dataclass

        """
        name: str


    @dataclass
    class Equipment:
        """
        Equipment dataclass

        """
        count: int


    @dataclass
    class NestedClasses:
        """
        NestedClasses dataclass

        """
        blah: str
        another_one: List['AnotherOne']
        just_something: int


    @dataclass
    class AnotherOne:
        """
        AnotherOne dataclass

        """
        testing: str
    '''

    gen_schema('test7.json')

    assert_py_code(expected, capfd)


def test_input_8(capfd):

    expected = '''
    from dataclasses import dataclass
    from typing import List, Optional, Union

    from dataclass_wizard import JSONWizard


    @dataclass
    class Container:
        """
        Container dataclass

        """
        data: 'Data'
        field_1: List['Data1']
        field_2: List['Data2']
        field_3: List['Data3']


    @dataclass
    class Data(JSONWizard):
        """
        Data dataclass

        """
        list_of_dictionaries: List['ListOfDictionary']


    @dataclass
    class ListOfDictionary:
        """
        ListOfDictionary dataclass

        """
        my_energies: List[Union['MyEnergy', int, str]]
        key: Optional[str]


    @dataclass
    class MyEnergy:
        """
        MyEnergy dataclass

        """
        my_test_val: Union[bool, int]
        another_val: str
        string_val: str
        merged_float: float


    @dataclass
    class Data1:
        """
        Data1 dataclass

        """
        key: str
        another_key: str


    @dataclass
    class Data2:
        """
        Data2 dataclass

        """
        question: str


    @dataclass
    class Data3:
        """
        Data3 dataclass

        """
        explanation: str
    '''

    gen_schema('test8.json')

    assert_py_code(expected, capfd)



from typing import Callable

from _typeshed import Incomplete

from ._type_def import DT as DT
from .utils._function_builder import FunctionBuilder as FunctionBuilder
from .utils._typing_compat import is_union as is_union

def process_patterned_date_time(func: Callable) -> Callable: ...
def _type_id(t) -> str: ...
def _generic_sig_str(name, args) -> str: ...
def _union_args(x): ...
def _flatten_union_args(args): ...
def _canonical_union_args(args): ...
def setup_recursive_safe_function(func: Callable = ..., *, fn_name: str | None = ..., is_generic: bool = ..., add_cls: bool = ..., prefix: str = ..., per_class_cache: bool = ...) -> Callable: ...
def setup_recursive_safe_function_for_generic(func: Callable = ..., prefix: str = ..., per_class_cache: bool = ...) -> Callable: ...

class cached_class_property:
    __func__: Callable
    __attr_name__: str
    def __init__(self, func) -> None: ...
    def __get__(self, instance, cls: Incomplete | None = ...): ...

class cached_property:
    __func__: Callable
    __attr_name__: str
    def __init__(self, func) -> None: ...
    def __get__(self, instance, cls: Incomplete | None = ...): ...



from .._bases_meta import DumpMeta
from .._dumpers import asdict
from .._lazy_imports import toml, toml_w
from .._loaders import fromdict, fromlist
from .._meta_cache import META_BY_DATACLASS


class TOMLWizard:
    # noinspection PyUnresolvedReferences,GrazieInspection
    """
    A Mixin class that makes it easier to interact with TOML data.

    .. NOTE::
      By default, *NO* key transform is used in the TOML dump process.
      In practice, this means that a `snake_case` field name in Python is saved
      as `snake_case` to TOML; however, this can easily be customized without
      the need to sub-class from :class:`JSONWizard`.

    For example:

        >>> @dataclass
        >>> class MyClass(TOMLWizard, dump_case='CAMEL'):
        >>>     ...

    """
    def __init_subclass__(cls, dump_case=None):
        """Allow easy setup of common config, such as key casing transform."""
        # Only add the key transform if Meta config has not been specified
        # for the dataclass.
        if dump_case and cls not in META_BY_DATACLASS:
            DumpMeta(case=dump_case).bind_to(cls)

    @classmethod
    def from_toml(cls,
                  string_or_stream, *,
                  decoder=None,
                  header='items',
                  parse_float=float):
        """
        Converts a TOML `string` to an instance of the dataclass, or a list of
        the dataclass instances.

        If ``header`` is provided and the corresponding value in the parsed
        data is a ``list``, the return type is ``List[T]``.
        """
        if decoder is None:  # pragma: no cover
            decoder = toml.loads

        o = decoder(string_or_stream, parse_float=parse_float)

        return (fromlist(cls, maybe_l)
                if (maybe_l := o.get(header)) and isinstance(maybe_l, list)
                else fromdict(cls, o))

    @classmethod
    def from_toml_file(cls, file, *,
                       decoder=None,
                       header='items',
                       parse_float=float):
        """
        Reads the contents of a TOML file and converts them
        into an instance (or list of instances) of the dataclass.

        Similar to :meth:`from_toml`, it can return a list if ``header``
        is specified and points to a list in the TOML data.
        """
        if decoder is None:  # pragma: no cover
            decoder = toml.load

        with open(file, 'rb') as in_file:
            return cls.from_toml(in_file,
                                 decoder=decoder,
                                 header=header,
                                 parse_float=parse_float)

    def to_toml(self,
                /,
                *encoder_args,
                encoder=None,
                multiline_strings=False,
                indent=4):
        """
        Converts a dataclass instance to a TOML `string`.

        Optional parameters include ``multiline_strings``
        for enabling/disabling multiline formatting of strings,
        and ``indent`` for setting the indentation level.
        """
        if encoder is None:  # pragma: no cover
            encoder = toml_w.dumps

        return encoder(asdict(self), *encoder_args,
                       multiline_strings=multiline_strings,
                       indent=indent)

    def to_toml_file(self, file, mode='wb',
                     encoder=None,
                     multiline_strings=False,
                     indent=4):
        """
        Serializes a dataclass instance and writes it to a TOML file.

        By default, opens the file in "write binary" mode.
        """
        if encoder is None:  # pragma: no cover
            encoder = toml_w.dump

        with open(file, mode) as out_file:
            self.to_toml(out_file, encoder=encoder,
                         multiline_strings=multiline_strings,
                         indent=indent)

    @classmethod
    def list_to_toml(cls,
                     instances,
                     header='items',
                     encoder=None,
                     **encoder_kwargs):
        """
        Serializes a ``list`` of dataclass instances into a TOML `string`,
        grouped under a specified header.
        """
        if encoder is None:
            encoder = toml_w.dumps

        list_of_dict = [asdict(o, cls=cls) for o in instances]

        return encoder({header: list_of_dict}, **encoder_kwargs)
